from gi.repository import Gtk

from yafti.abc import YaftiScreen

_xml = """\
<?xml version="1.0" encoding="UTF-8"?>
<interface>
    <requires lib="gtk" version="4.0"/>
    <template class="YaftiConsoleScreen" parent="GtkScrolledWindow">
        <property name="margin-start">40</property>
        <property name="margin-end">40</property>
        <property name="margin-bottom">40</property>
        <property name="height-request">350</property>
        <child>
            <object class="GtkBox" id="console_output">
                <property name="visible">True</property>
                <property name="margin-top">12</property>
                <property name="margin-start">12</property>
                <property name="margin-end">12</property>
                <property name="orientation">vertical</property>
            </object>
        </child>
        <style>
            <class name="card"/>
        </style>
    </template>
</interface>
"""


@Gtk.Template(string=_xml)
class ConsoleScreen(YaftiScreen, Gtk.ScrolledWindow):
    __gtype_name__ = "YaftiConsoleScreen"

    console_output = Gtk.Template.Child()

    def stdout(self, text):
        if isinstance(text, bytes):
            t = text.decode()
            for line in t.split("\n"):
                if not line:
                    continue
                self.stdout(Gtk.Text(text=line))
        else:
            self.console_output.append(text)
        self.scroll_to_bottom()

    def stderr(self, text):
        if isinstance(text, bytes):
            t = text.decode()
            for line in t.split("\n"):
                if not line:
                    continue
                self.stderr(Gtk.Text(text=line))
        else:
            self.console_output.append(text)
        self.scroll_to_bottom()

    def scroll_to_bottom(self):
        adj = self.get_vadjustment()
        adj.set_value(adj.get_upper())
        self.set_vadjustment(adj)

    def scroll_to_top(self):
        adj = self.get_vadjustment()
        adj.set_value(adj.get_lower())

    def hide(self):
        self.set_visible(False)

    def show(self):
        self.set_visible(True)

    def toggle_visible(self):
        self.set_visible(self.get_visible() is False)



"""
Copyright 2023 Marco Ceppi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

/f

Run a command on the system

Configuration usage example:

  commands:
    pre:
    # Simple config
    - run: /usr/bin/whoami
    # Explicit full path plugin
    - yafti.plugin.run: /usr/bin/whoami
    # Run with parameters
    - run: /bin/ls -lah
    - run: ["/bin/ls", "-lah"]
    - run:
      - /bin/ls
      - "-lah"


Programmatic usage example:

  from yafti.plugin.run import Run
  r = Run()
  r.exec(["/usr/bin/whoami"])
  f.exec(pkg="com.github.marcoceppi.PackageName", reinstall=True)

  r("/usr/bin/whoami")
  r(cmd="/usr/bin/whoami")
  r(cmd=["/usr/bin/whoami"])

"""

import asyncio
import shlex
import subprocess
from os.path import isfile
from shutil import which

from pydantic import validate_call

from yafti import log
from yafti.abc import YaftiPlugin, YaftiPluginReturn


class Run(YaftiPlugin):
    async def exec(self, cmd: str) -> subprocess.CompletedProcess:
        log.debug("running command", cmd=cmd)

        # spawn command in host when running from container
        is_container = isfile("/run/.containerenv") or isfile("/.dockerenv")
        if not isfile(cmd) and is_container:
            if which("distrobox-host-exec"):
                cmd = f"distrobox-host-exec {cmd}"
            elif which("flatpak-spawn"):
                cmd = f"flatpak-spawn --host {cmd}"

        proc = await asyncio.create_subprocess_shell(
            cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        stdout, stderr = await proc.communicate()

        log.info("command complete", cmd=cmd, code=proc.returncode)
        log.debug(
            "command complete",
            cmd=cmd,
            code=proc.returncode,
            stdout=stdout,
            stderr=stderr,
        )

        return subprocess.CompletedProcess(
            cmd, returncode=proc.returncode, stdout=stdout, stderr=stderr
        )

    async def install(self, package: str) -> YaftiPluginReturn:
        """Execute a command on the host system

        Args:
          package: The command to execute

        Returns:
          An object containing the stdout and stderr from the command
        """
        return await self.exec(package)

    @validate_call
    async def __call__(self, cmd: list[str] | str) -> YaftiPluginReturn:
        log.debug("run called", cmd=cmd)
        if isinstance(cmd, list):
            cmd = shlex.join(cmd)

        r = await self.exec(cmd)
        return YaftiPluginReturn(output=r.stdout, errors=r.stderr, code=r.returncode)



from functools import partial
import json

from gi.repository import Adw, Gtk
from pydantic import BaseModel

from yafti import log
from yafti.abc import YaftiScreen
from yafti.screen.dialog import DialogBox
from yafti.screen.package.state import PackageScreenState
from yafti.screen.utils import find_parent

_xml = """\
<?xml version="1.0" encoding="UTF-8"?>
<interface>
  <requires lib="gtk" version="4.0"/>
  <template class="YaftiPackagePickerScreen" parent="AdwBin">
    <property name="halign">fill</property>
    <property name="valign">fill</property>
    <property name="hexpand">true</property>
    <child>
      <object class="AdwStatusPage" id="status_page">
        <property name="halign">fill</property>
        <property name="valign">fill</property>
        <property name="hexpand">true</property>
        <child>
          <object class="GtkBox">
            <property name="orientation">vertical</property>
            <property name="vexpand">true</property>
            <property name="hexpand">true</property>
            <property name="valign">center</property>
            <child>
              <object class="AdwPreferencesPage">
                <child>
                  <object class="AdwPreferencesGroup" id="package_list"></object>
                </child>
              </object>
            </child>
          </object>
        </child>
      </object>
    </child>
  </template>
</interface>
"""


@Gtk.Template(string=_xml)
class PackagePickerScreen(YaftiScreen, Adw.Bin):
    __gtype_name__ = "YaftiPackagePickerScreen"

    status_page = Gtk.Template.Child()
    package_list = Gtk.Template.Child()

    class Config(BaseModel):
        title: str = "Package Installation"
        packages: list | dict

    def __init__(
        self,
        state: PackageScreenState,
        title: str,
        packages: list | dict,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.status_page.set_title(title)
        self.packages = packages
        self.state = state
        self.draw()

    def draw(self):
        if isinstance(self.packages, list):
            for item in self._build_apps(self.packages):
                self.package_list.add(item)
            return

        for name, details in self.packages.items():
            action_row = Adw.ActionRow(title=name, subtitle=details.get("description"))

            def state_set(group, _, value):
                self.state.set(f"group:{group}", value)
                d = self.packages.get(group)
                for pkg in d.get("packages", []):
                    for pkg_name in pkg.values():
                        if isinstance(pkg_name, dict):
                            pkg_name = json.dumps(pkg_name)
                        self.state.set(f"pkg:{pkg_name}", value)

            state_set(name, None, details.get("default", True))
            _switcher = Gtk.Switch()
            _switcher.set_active(self.state.get(f"group:{name}"))
            _switcher.set_valign(Gtk.Align.CENTER)

            state_set_fn = partial(state_set, name)
            _switcher.connect("state-set", state_set_fn)
            action_row.add_suffix(_switcher)

            _customize = Gtk.Button()
            _customize.set_icon_name("go-next-symbolic")
            _customize.set_valign(Gtk.Align.CENTER)
            _customize.add_css_class("flat")
            action_row.add_suffix(_customize)
            picker_fn = partial(self._build_picker, details.get("packages", []))
            _customize.connect("clicked", picker_fn)
            self.package_list.add(action_row)

    def _build_picker(self, packages: list, *args):
        dialog = DialogBox(find_parent(self, Gtk.Window))

        btn_cancel = Gtk.Button()
        btn_save = Gtk.Button()
        btn_cancel.set_label("Cancel")
        btn_save.set_label("Save")
        btn_save.add_css_class("suggested-action")

        header = Adw.HeaderBar()
        header.pack_start(btn_cancel)
        header.pack_end(btn_save)
        header.set_show_end_title_buttons(False)
        header.set_show_start_title_buttons(False)

        item_list = Adw.PreferencesGroup()
        item_list.set_description(
            "The following list includes only applications available in your preferred "
            "package manager."
        )
        page = Adw.PreferencesPage()
        page.add(item_list)

        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.append(header)
        box.append(page)

        dialog.set_content(box)
        dialog.set_default_size(500, 600)

        for item in self._build_apps(packages):
            item_list.add(item)

        btn_cancel.connect("clicked", lambda x: dialog.close())
        btn_save.connect("clicked", lambda x: dialog.close())
        dialog.show()

    def _build_apps(self, packages: list):
        for item in packages:
            for name, pkg in item.items():
                _apps_action_row = Adw.ActionRow(
                    title=name,
                )
                _app_switcher = Gtk.Switch()
                if isinstance(pkg, dict):
                    pkg = json.dumps(pkg)
                _app_switcher.set_active(self.state.get(f"pkg:{pkg}"))
                _app_switcher.set_valign(Gtk.Align.CENTER)

                def set_state(pkg, btn, value):
                    log.debug("state-set", pkg=pkg, value=value)
                    self.state.set(f"pkg:{pkg}", value)

                set_state_func = partial(set_state, pkg)
                _app_switcher.connect("state-set", set_state_func)
                _apps_action_row.add_suffix(_app_switcher)
                yield _apps_action_row



from pydantic import RootModel, BaseModel


PackageConfig = RootModel[dict[str, str | dict]]


class PackageGroupConfigDetails(BaseModel):
    description: str
    default: bool = True
    packages: list[PackageConfig]


PackageGroupConfig = RootModel[dict[str, PackageGroupConfigDetails]]



import json
from typing import Any
from hashlib import sha256


def generate_fingerprint(obj: Any):
    return sha256(json.dumps(obj).encode()).hexdigest()


def parse_packages(packages: dict | list) -> dict:
    output = {}

    if isinstance(packages, dict):
        for group, value in packages.items():
            output[f"group:{group}"] = True
            output.update(parse_packages(value["packages"]))
        return output

    for pkgcfg in packages:
        for package in pkgcfg.values():
            if isinstance(package, dict):
                package = json.dumps(package)
            output[f"pkg:{package}"] = True
    return output



import asyncio
from functools import partial

from gi.repository import Adw, Gtk

import yafti.share
from yafti import events
from yafti.registry import SCREENS

_xml = """\
<?xml version="1.0" encoding="UTF-8"?>
<interface>
  <requires lib="gtk" version="4.0"/>
  <requires lib="libadwaita" version="1.0" />
  <template class="YaftiWindow" parent="AdwApplicationWindow">
    <property name="default-width">750</property>
    <property name="default-height">640</property>
    <property name="title">Welcome!</property>
    <child>
      <object class="GtkBox">
        <property name="orientation">vertical</property>
        <child>
          <object class="AdwHeaderBar" id="headerbar">
            <style>
              <class name="flat" />
            </style>
            <property name="title_widget">
              <object class="AdwCarouselIndicatorDots" id="carousel_indicator">
                <property name="carousel">carousel</property>
                <property name="orientation">horizontal</property>
              </object>
            </property>
            <property name="show-end-title-buttons">False</property>
            <child type="start">
              <object class="GtkButton" id="btn_back">
                <property name="label" translatable="yes">Back</property>
                <property name="halign">center</property>
                <property name="visible">False</property>
              </object>
            </child>
            <child type="end">
              <object class="GtkButton" id="btn_next">
                <property name="label" translatable="yes">Next</property>
                <property name="halign">center</property>
                <property name="visible">True</property>
                <style>
                  <class name="suggested-action" />
                </style>
              </object>
            </child>
          </object>
        </child>
        <child>
          <object class="AdwToastOverlay" id="toasts">
            <child>
              <object class="AdwCarousel" id="carousel">
                <property name="vexpand">True</property>
                <property name="hexpand">True</property>
                <property name="allow_scroll_wheel">False</property>
                <property name="allow_mouse_drag">False</property>
                <property name="allow_long_swipes">False</property>
              </object>
            </child>
          </object>
        </child>
      </object>
    </child>
  </template>
</interface>
"""


@Gtk.Template(string=_xml)
class Window(Adw.ApplicationWindow):
    __gtype_name__ = "YaftiWindow"

    carousel_indicator = Gtk.Template.Child()
    carousel = Gtk.Template.Child()
    headerbar = Gtk.Template.Child()
    btn_back = Gtk.Template.Child()
    btn_next = Gtk.Template.Child()
    toasts = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.app = kwargs.get("application")
        events.register("btn_next")
        events.register("btn_back")
        events.on("btn_next", self.next)
        events.on("btn_back", self.back)

        # TODO(GH-2): not a huge fan of this
        yafti.share.BTN_BACK = self.btn_back
        yafti.share.BTN_NEXT = self.btn_next

        def do_emit(*args, **kwargs):
            asyncio.create_task(events.emit(*args, **kwargs))

        _next = partial(do_emit, "btn_next")
        _back = partial(do_emit, "btn_back")

        self.connect("show", self.draw)
        self.connect("close-request", self.app.quit)
        self.btn_next.connect("clicked", _next)
        self.btn_back.connect("clicked", _back)
        self.carousel.connect("page-changed", self.changed)

    def draw(self, _) -> None:
        asyncio.ensure_future(self.build_screens())

    async def build_screens(self):
        screens = self.app.config.screens
        for name, details in screens.items():
            if details.source not in SCREENS:
                continue
            screen = SCREENS.get(details.source)
            s = await screen.from_config(details.values)
            if s is None:
                continue
            self.carousel.append(s)

    @property
    def idx(self) -> float:
        return self.carousel.get_position()

    def goto(self, page: int, animate: bool = True) -> None:
        if page < 0:
            page = 0

        if page >= self.carousel.get_n_pages():
            page = self.carousel.get_n_pages()

        current_screen = self.carousel.get_nth_page(self.idx)
        next_screen = self.carousel.get_nth_page(page)

        current_screen.deactivate()
        self.carousel.scroll_to(next_screen, animate)

    @property
    def is_last_page(self):
        return self.idx + 1 >= self.carousel.get_n_pages()

    async def next(self, _) -> None:
        if self.idx + 1 >= self.carousel.get_n_pages():
            self.app.quit()
        else:
            self.goto(self.idx + 1)

    async def back(self, _) -> None:
        self.goto(self.idx - 1)

    def changed(self, *args) -> None:
        self.btn_back.set_visible(self.idx > 0)
        current_screen = self.carousel.get_nth_page(self.idx)
        if self.idx + 1 >= self.carousel.get_n_pages():
            self.btn_next.set_label("Done")
            self.btn_back.set_visible(False)
        else:
            self.btn_next.set_label("Next")
        current_screen.activate()



import logging

__all__ = ["info", "warn", "error", "debug", "set_level"]

_l = logging.getLogger("yafti")


def _fmt(msg: dict) -> str:
    return " ".join([f"{k}={v}" for k, v in msg.items()])


def set_level(level):
    _l.setLevel(level)


def debug(message, **kwargs):
    _l.debug(f"{message} {_fmt(kwargs)}")


def info(message, **kwargs):
    _l.info(f"{message} {_fmt(kwargs)}")


def warn(message, *kwargs):
    _l.warn(f"{message} {_fmt(kwargs)}")


def error(message, *kwargs):
    _l.error(f"{message} {_fmt(kwargs)}")
